/* -*- linux-c -*-
 * FILE NAME esdcan.c
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *	     This file contains the Linux / RTAI-Linux entries for
 *           the esd CAN driver
 *
 *
 * Author:   Matthias Fuchs
 *           matthias.fuchs@esd-electronics.com
 *
 * history:
 *
 *  $Log: esdcan.c,v $
 *  Revision 1.109  2006/10/12 09:53:36  andreas
 *  Replaced some forgotten CANIO_-error codes with the OSIF_ counterparts
 *  Replaces some errnos with their OSIF_counterparts
 *  Cleaned up return of error codes to user space,
 *    please use positive error codes, only!!!
 *
 *  Revision 1.108  2006/10/11 10:12:48  andreas
 *  Fixed compilation problem with SuSE > 10.0
 *
 *  Revision 1.107  2006/08/17 13:26:59  michael
 *  Rebirth of OSIF_ errorcodes
 *
 *  Revision 1.106  2006/07/11 15:13:58  manuel
 *  Replaced nuc_timestamp with data from CAN_STAT structure in can_read_proc
 *
 *  Revision 1.105  2006/07/04 12:41:05  andreas
 *  Added missing IOCTL_ESDCAN_SET_ALT_RTR_ID
 *
 *  Revision 1.104  2006/06/29 11:53:17  andreas
 *  Added module load parameter "errorinfo" to disable extended error info
 *
 *  Revision 1.103  2006/06/27 13:12:43  andreas
 *  Exchanged OSIF_errors with CANIO_errors
 *
 *  Revision 1.102  2006/06/27 10:00:45  andreas
 *  Removed usage of nuc_baudrate_get()
 *  Moved baud_table into nucleus
 *  Added TODO about timestamps on baudrate change events
 *
 *  Revision 1.101  2006/04/28 12:30:36  andreas
 *  Added warning in syslog, if debug driver
 *  Changes to support recent kernel (>2.6.13)
 *  Added usage of compat_ioctl and unlocked_ioctl
 *
 *  Revision 1.100  2006/02/14 09:20:52  matthias
 *  Don't use macros for esdcan_un/register_ioctl32 and others when these functions are not used. USe empty functions instead. This avoid compiler warnings from gcc 4.x
 *
 *  Revision 1.99  2005/12/08 11:35:26  andreas
 *  For kernels below 2.4.0 legacy PCI support is used instead of hotplugging
 *
 *  Revision 1.98  2005/12/06 13:26:10  andreas
 *  No functional changes. Removed old "#if 0" path and small cleanup.
 *
 *  Revision 1.97  2005/11/14 13:45:30  manuel
 *  Added #if PLATFORM_64BIT around esdcan_register_ioctl32 and esdcan_unregister_ioctl32 prototypes
 *
 *  Revision 1.96  2005/11/02 07:12:46  andreas
 *  Added IOCTL_ESDCAN_SET_20B_HND_FILTER (de-)registration for 64-Bit platforms
 *
 *  Revision 1.95  2005/10/06 07:39:17  matthias
 *  temporarily disabled nuc_timestamp call for proc interface
 *
 *  Revision 1.94  2005/08/29 14:33:48  andreas
 *  Fixed bug in release of memory regions
 *
 *  Revision 1.93  2005/07/29 08:19:43  andreas
 *  crd-structure stores pointer (pCardIdent) into cardFlavours structure instead of index (flavour), now
 *
 *  Revision 1.92  2005/07/28 08:07:30  andreas
 *  Separated initialization of different board types (pci, raw, usb,...).
 *  From now on this file contains function common to all (or at least several)
 *  board types. Everything specific to one board type is located in esdcan_pci.c,
 *  esdcan_raw.c, esdcan_usb.c or esdcan_pci_legacy.c.
 *  BOARD and BOARD_NAME defines have been removed. There's a ESDCAN_DRIVER_NAME defined
 *  in each boardrc.h.
 *  There's also an array cardFlavours defined in boardrc.c, which provides the
 *  "spaces" and "irq" arrays, board name and type specific IDs for all boards supported
 *  by the same driver binary. This means at least for PCI and hotplug PCI the driver
 *  can support several different devices, now.
 *  Bugfixes in ISA-initialization (behaviour, if more than one address space is defined),
 *  located in esdcan_raw.c, now.
 *
 *  Revision 1.91  2005/04/25 07:29:49  andreas
 *  Fixed bug (irrelevant, since feature isn't used until now) in ISA module paramters
 *
 *  Revision 1.90  2005/04/20 13:44:20  andreas
 *  Changed for 64-Bit Linux (ioctl registration, C-types)
 *  Cleanup
 *
 *  Revision 1.89  2005/03/04 08:41:16  andreas
 *  Added cpci200
 *
 *  Revision 1.88  2005/02/21 15:24:12  matthias
 *  -because CM-size increased we have to copy CMSG and CMSG_T struct from/to userspace
 *  -added pci405fw2
 *  -removed unused code
 *
 *  Revision 1.87  2004/11/15 13:29:52  matthias
 *  added cpci750 support
 *
 *  Revision 1.86  2004/11/10 14:33:03  michael
 *  Compilable with versioned symbols again
 *
 *  Revision 1.85  2004/09/23 14:49:32  andreas
 *  forgot to remove verbose-preset
 *
 *  Revision 1.84  2004/09/23 14:48:17  andreas
 *  Moved most of the USB code to esdcan_usb.c
 *  Changed debug output to dprint
 *
 *  11.05.04 - Added OSIF_CALLTYPE, where needed
 *             Removed warning (passing param 4 of nuc_extra_info)  ab
 *  07.05.04 - Changes for kernel 2.6:
 *               - module-parameters (and description)
 *               - module open count
 *               - check_region -> request_region
 *               - EXPORT_NO_SYMBOLS removed
 *             Some cleanup                                         ab
 *  07.04.04 - changed major of gitane to 54                        ab
 *  11.03.04 - added Gitane and GitaneLite majors
 *             added IO-space mapping for PCI-cards                 ab
 *  17.12.03 - added pci266                                         ab
 *  01.12.03 - resetting USB-modules on driver-unload
 *             some more small corrections for USB-modules          ab
 *  26.11.03 - ifdefed release_region for kernel < 2.4.0            ab
 *  13.11.03 - USB331-Support added (THIS IS STILL BETA!!!)
 *             USB-specific stuff is implemented here, to be able
 *             reimplement USB-stuff for other systems.
 *             There are NO effects for non-USB-cards. Nevertheless
 *             some cleanup is still needed.                        ab
 *  05.11.03 - Added support for four subvendor and subsystem-IDs   ab
 *  27.10.03 - Corrected usage of print-macros
 *             Added verbose module parameter                       ab
 *  29.09.03 - read/write: get node-mutex before common call
 *             (same as IRIX IOCTRL)                                sr
 *  05.09.03 - improved module parameter handling for irq and io
 *             (only for RAW_BOARDS)
 *             a) irq arg can be terminated by -1: irq=25,-1
 *             b) io arg can be temrinated by 0: io=0xf0000000,0    mf
 *  03.06.03 - undo change from 21.05.03                            mf
 *  21.05.03 - release irq when nuc_node_attach failed (0.3.3)      mf
 *  15.01.03 - common driver part extracted "../esdcan_common.c"    sr
 *  18.11.02 - added callback interface for nucleus                 mf
 *  07.11.02 - Version 0.1.1                                        mf
 *  21.05.02 - added hardware timestamping (Version 0.1.0)          mf
 *  17.05.02 - first release (Version 0.1.0)                        mf
 *  02.05.02 - first version                                        mf
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file esdcan.c
    \brief Linux driver entries

    This file contains the Linux entries for
    the esd CAN driver.
*/

#include <linux/config.h>
#include <linux/version.h>
#ifdef DISTR_SUSE
# undef CONFIG_MODVERSIONS
#endif
#ifdef CONFIG_MODVERSIONS
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#  include <config/modversions.h>
# else
#  include <linux/modversions.h>
# endif
# ifndef MODVERSIONS
#  define MODVERSIONS
# endif
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#include <linux/ioctl32.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <linux/ioport.h>
#endif
#ifndef OSIF_OS_RTAI
#include <asm/uaccess.h>
#endif
#ifdef LTT
#warning "Tracing enabled"
#include <linux/trace.h>
#endif

#include <osif.h>
#include <cm.h>
#include <boardrc.h>
#include <board.h>
#include <nucleus.h>
#include <esdcan.h>
#include <canio.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
# if defined(BOARD_PCI)
#  define ESDCAN_PCI_HOTPLUG
#  undef BOARD_PCI /* Disable legacy PCI support */
# endif
#endif

#if defined(BOARD_USB) || defined(ESDCAN_PCI_HOTPLUG)
# define ESDCAN_HOTPLUG_DRIVER
#endif

#if defined (BOARD_USB)
# if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
#  include <linux/usb.h>
# endif
#endif

#ifdef ESDDBG
# define CAN_DEBUG
#endif

#ifdef CAN_DEBUG
# define CAN_DBG(fmt)	    OSIF_DPRINT(fmt)  /*!< Macro for debug prints */
# define CAN_IRQ_DBG(fmt)   OSIF_DPRINT(fmt)  /*!< Macro for debug prints on interrupt-level */
#else
# define CAN_DBG(fmt)                         /**< Empty expression */
# define CAN_IRQ_DBG(fmt)                     /**< Empty expression */
#endif
#define CAN_PRINT(fmt)	    OSIF_PRINT(fmt)   /**< Makro for output to the general user (NOT surpressed in release version) */
#define CAN_DPRINT(fmt)	    OSIF_DPRINT(fmt)  /**< Makro for output to the general user (filtered by verbose level)         */
#define CAN_IRQ_PRINT(fmt)  OSIF_DPRINT(fmt)  /**< Makro for output to the general user on interrupt level (NOT surpressed in release version) */

#define ESDCAN_ZONE_INIFU    OSIF_ZONE_ESDCAN|OSIF_ZONE_INIT|OSIF_ZONE_FUNC
#define ESDCAN_ZONE_INI      OSIF_ZONE_ESDCAN|OSIF_ZONE_INIT
#define ESDCAN_ZONE_FU       OSIF_ZONE_ESDCAN|OSIF_ZONE_FUNC
#define ESDCAN_ZONE          OSIF_ZONE_ESDCAN
#define ESDCAN_ZONE_IRQ      OSIF_ZONE_IRQ    /* DO NOT USE ANY MORE BITS (nto will thank you:) */
#define ESDCAN_ZONE_ALWAYS   0xFFFFFFFF
#define ESDCAN_ZONE_WARN     OSIF_ZONE_ESDCAN|OSIF_ZONE_WARN
#define ESDCAN_ZONE_USR_INIT OSIF_ZONE_ESDCAN|OSIF_ZONE_USR_INIT

#ifdef LTT
INT32 can_ltt0;
UINT8 can_ltt0_byte;
#endif

/* module parameter verbose: default changes regarding to debug flag */
#ifndef CAN_DEBUG
/*
 *  The released driver writes load- and unload-messages to syslog,
 *  if the user hasn't specified a verbose level.
 */
uint verbose = 0x00000001;
#else
/*
 *  The debug-version of the driver writes all "release-version"-output
 *  and warning- and error-messages, if the developer doesn't specify
 *  otherwise.
 */
/* uint verbose = 0xF8F00001; *//* AB setting */
/* uint verbose = 0xF7FFFFFF; *//* AB all but IRQ */
uint verbose = 0xC00000FF;
#endif

unsigned int major = MAJOR_LINUX; /* MAJOR_LINUX is set in .cfg files */

VOID esdcan_show_card_info( CAN_CARD *crd );
VOID esdcan_show_card_info_all( VOID );
VOID esdcan_show_driver_info( VOID );
INT32 can_attach_common(CAN_CARD *crd);

CAN_NODE *nodes[MAX_CARDS * NODES_PER_CARD];  /* index: minor */

int mode = 0;
int baud = CANIO_BAUD_NOT_SET;
int errorinfo = 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/* common module parameter */
module_param(major, uint, 0);
MODULE_PARM_DESC(major, "major number to be used for the CAN card");
module_param(mode, int, 0);     /* e.g. LOM */
MODULE_PARM_DESC(mode, "activate certain driver modes, e.g. LOM");
module_param(baud, int, 0);     /* initial baudrate */
MODULE_PARM_DESC(baud, "change initial baudrate of CAN card");
module_param(verbose, uint, 0); /* mask for output selection */
MODULE_PARM_DESC(verbose, "change verbose level of driver");
module_param(errorinfo, int, 0); /* flag for extended error info */
MODULE_PARM_DESC(errorinfo, "enable/disable extended error info (default: on)");
unsigned int  compat32 = 1;
module_param(compat32, int, 0); /* setting this to zero disables 32-bit compatibility on 64-bit systems */
MODULE_PARM_DESC(compat32, "disable 32-Bit compatibility on 64-Bit systems");
#else
/* common module parameter */
MODULE_PARM(major, "1i");
MODULE_PARM(mode, "1i");    /* e.g. LOM */
MODULE_PARM(baud, "1i");    /* initial baudrate */
MODULE_PARM(verbose, "1i"); /* mask for output selection */
MODULE_PARM(errorinfo, "1i"); /* enable/disable extended error info (default: on) */
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
# define INODE2CANNODE(inode) (iminor(inode))
#else
# define INODE2CANNODE(inode) (MINOR(inode->i_rdev) & 0xff)
#endif

#ifdef OSIF_OS_LINUX
#define ESDCAN_IOCTL_PROTO \
        static int esdcan_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
# if (PLATFORM_64BIT) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static long esdcan_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
# endif
# if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
static long esdcan_unlocked_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
# endif
#endif /* OSIF_OS_LINUX */

#ifdef OSIF_OS_RTAI
#define ESDCAN_IOCTL_PROTO \
        INT32 esdcan_ctrl(INT32 hnd, UINT32 cmd, UINT32 arg)
#endif /* OSIF_OS_RTAI */

static int esdcan_open(struct inode *inode, struct file *file);
static int esdcan_release(struct inode *inode, struct file *file);
ESDCAN_IOCTL_PROTO;

void esdcan_unregister_ioctl32(void);
int32_t esdcan_register_ioctl32(void);

#define RETURN_TO_USER(ret) return(-ret)


#ifndef OSIF_OS_RTAI
struct file_operations esdcan_fops =
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	esdcan_ioctl,
	NULL,
	esdcan_open,
	NULL,
	esdcan_release,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	owner:	 THIS_MODULE,
	ioctl:	 esdcan_ioctl,
	open:	 esdcan_open,
	release: esdcan_release,
#else
	.owner =   THIS_MODULE,
	.ioctl =   esdcan_ioctl,
# if (PLATFORM_64BIT) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
        .compat_ioctl = esdcan_compat_ioctl,
# endif
# if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
        .unlocked_ioctl = esdcan_unlocked_ioctl,
# endif
	.open =    esdcan_open,
	.release = esdcan_release,
#endif /*  LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0) */
};
#endif /* OSIF_OS_RTAI */


#define  PROCFN_LEN 30
#if (defined BUS_STATISTICS) && ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) )
UINT8  procfn_bus_statistic[PROCFN_LEN];

int can_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	CAN_NODE **nodes = (CAN_NODE**)data;
	int len = 0;
	int n;

	for (n = 0; n < MAX_CARDS * NODES_PER_CARD; n++) {
		CAN_NODE *node = nodes[n];
		CAN_STAT stat;
		int l;
		if ( node ) {
			nuc_statistics( node, &stat );
			len += OSIF_SNPRINTF((buf+len, count-len, "node%d %u ", n, stat.baud));
			for (l=0; l<=8; l++)
				len += OSIF_SNPRINTF((buf+len, count-len, "%u ", stat.msg_count_a[l]));
			for (l=0; l<=8; l++)
				len += OSIF_SNPRINTF((buf+len, count-len, "%u ", stat.msg_count_b[l]));
			len += OSIF_SNPRINTF((buf+len, count-len, "%u ", stat.byte_count));
			len += OSIF_SNPRINTF((buf+len, count-len, "%llu %llu\n", stat.ts.tick, stat.ts_freq.tick));
		}
	}
	*eof = 1;
	return len;
}

void esdcan_proc_bus_statistic_create(void)
{
	CAN_DBG((ESDCAN_ZONE_INI, "%s: creating proc entry (bus/can/%s)\n",
                 OSIF_FUNCTION, ESDCAN_DRIVER_NAME));

	/* create /proc/bus/can/<BOARD> */
	OSIF_SNPRINTF((procfn_bus_statistic, PROCFN_LEN,"bus/can/%s", ESDCAN_DRIVER_NAME));
	if (!create_proc_read_entry( procfn_bus_statistic,
				     0 /* default mode */,
				     NULL /* parent dir */,
				     can_read_proc,
				     (VOID*)nodes /* arg */)) {
		proc_mkdir("bus/can", NULL);
		create_proc_read_entry( procfn_bus_statistic,
					0 /* default mode */,
					NULL /* parent dir */,
					can_read_proc,
					(VOID*)nodes /* arg */);
	}
}
void esdcan_proc_bus_statistic_remove(void)
{
	CAN_DBG((ESDCAN_ZONE_INI, "%s: removing proc entry bus statistic\n", OSIF_FUNCTION));
	remove_proc_entry(procfn_bus_statistic, NULL);
}
#else
void esdcan_proc_bus_statistic_create(void) {}
void esdcan_proc_bus_statistic_remove(void) {}
#endif

#if (defined PROC_EXTRA_INFO) && ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) )
UINT8  procfn_extra_info[PROCFN_LEN];

int can_read_proc_extrainfo(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	CAN_NODE **nodes = (CAN_NODE**)data;
	INT32 len = 0;
	int n;

	for (n = 0; n < MAX_CARDS * NODES_PER_CARD; n++) {
		CAN_NODE *node = nodes[n];
		if ( node ) {
			len += OSIF_SNPRINTF((buf+len, count-len, "--- net%d ---\n", node->net_no));
			nuc_extra_info(node, buf+len, count-len, &len);
		}
	}
	*eof = 1;
	return len;
}

void esdcan_proc_extra_info_create(void)
{
	CAN_DBG((ESDCAN_ZONE_INI, "%s: creating proc entry (bus/can/%s_ex)\n",
                 OSIF_FUNCTION, ESDCAN_DRIVER_NAME));

	/* create /proc/bus/can/<BOARD> */
	OSIF_SNPRINTF((procfn_extra_info, PROCFN_LEN,"bus/can/%s_ex",
                       ESDCAN_DRIVER_NAME));
	if (!create_proc_read_entry( procfn_extra_info,
				     0 /* default mode */,
				     NULL /* parent dir */,
				     can_read_proc_extrainfo,
				     (VOID*)nodes /* arg */)) {
		proc_mkdir("bus/can", NULL);
		create_proc_read_entry( procfn_extra_info,
					0 /* default mode */,
					NULL /* parent dir */,
					can_read_proc_extrainfo,
					(VOID*)nodes /* arg */);
	}
}
void esdcan_proc_extra_info_remove(void)
{
	CAN_DBG((ESDCAN_ZONE_INI, "%s: removing proc entry extra info\n", OSIF_FUNCTION));
	remove_proc_entry(procfn_extra_info, NULL);
}
#else
void esdcan_proc_extra_info_create(void) {}
void esdcan_proc_extra_info_remove(void) {}
#endif

#if (PLATFORM_64BIT) && (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10))
static int esdcan_ioctl_32(unsigned int fd, unsigned int cmd, unsigned long arg, struct file *file);

void esdcan_unregister_ioctl32(void)
{
        if (compat32) {
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_CREATE);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_ID_RANGE_ADD);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_ID_RANGE_DEL);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SET_BAUD);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SET_TIMEOUT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SEND);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TAKE);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_DEBUG);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_BAUD);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_DESTROY);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_DESTROY_DEPRECATED);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_ABORT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_RX_ABORT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_CREATE);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_UPDATE);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_OFF);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_DESTROY);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_RX_OBJ);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_READ);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_WRITE);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_UPDATE);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_FLUSH_RX_FIFO);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_FLUSH_RX_FIFO_DEPRECATED);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_RX_MESSAGES);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_RX_TIMEOUT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TX_TIMEOUT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TIMESTAMP_FREQ);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TIMESTAMP);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TICKS_FREQ);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TICKS);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SEND_T);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_WRITE_T);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TAKE_T);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_READ_T);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_PURGE_TX_FIFO);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_PURGE_TX_FIFO_DEPRECATED);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_ON);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SET_RX_TIMEOUT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SET_TX_TIMEOUT);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_FEATURES);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_SCHEDULE);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_TX_OBJ_SCHEDULE_START);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_TX_OBJ_SCHEDULE_STOP);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_SET_20B_HND_FILTER);
                UNREGISTER_IOCTL_32(IOCTL_ESDCAN_GET_SERIAL);
                UNREGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_SET_ALT_RTR_ID);
                UNREGISTER_IOCTL_COMPAT(IOCTL_CAN_SET_QUEUESIZE); /* special ioctl (from candev driver) for driver version detection */
        }
}

int32_t esdcan_register_ioctl32(void)
{
        int32_t result = 0;
        if (compat32) {
                /* register 32-bit ioctls */
                REGISTER_IOCTL_32(IOCTL_ESDCAN_CREATE, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_ID_RANGE_ADD, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_ID_RANGE_DEL, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SET_BAUD, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SET_TIMEOUT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SEND, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TAKE, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_DEBUG, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_BAUD, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_DESTROY, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_DESTROY_DEPRECATED, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_ABORT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_RX_ABORT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_CREATE, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_UPDATE, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_OFF, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_DESTROY, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_RX_OBJ, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_READ, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_WRITE, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_UPDATE, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_FLUSH_RX_FIFO, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_FLUSH_RX_FIFO_DEPRECATED, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_RX_MESSAGES, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_RX_TIMEOUT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TX_TIMEOUT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TIMESTAMP_FREQ, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TIMESTAMP, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TICKS_FREQ, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_TICKS, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SEND_T, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_WRITE_T, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TAKE_T, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_READ_T, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_PURGE_TX_FIFO, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_PURGE_TX_FIFO_DEPRECATED, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_ON, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SET_RX_TIMEOUT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SET_TX_TIMEOUT, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_FEATURES, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_TX_OBJ_SCHEDULE, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_TX_OBJ_SCHEDULE_START, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_TX_OBJ_SCHEDULE_STOP, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_SET_20B_HND_FILTER, result);
                REGISTER_IOCTL_32(IOCTL_ESDCAN_GET_SERIAL, result);
                REGISTER_IOCTL_COMPAT(IOCTL_ESDCAN_SET_ALT_RTR_ID, result);
                REGISTER_IOCTL_COMPAT(IOCTL_CAN_SET_QUEUESIZE, result); /* special ioctl (from candev driver) for driver version detection */
                if ( 0 != result ) {
                        CAN_PRINT(("esd CAN driver: An ioctl collides with an ioctl of another!\n"));
                        CAN_PRINT(("esd CAN driver: driver on your system!\n"));
                        CAN_PRINT(("esd CAN driver: Since this is a problem of the 32-bit compatibility\n"));
                        CAN_PRINT(("esd CAN driver: layer of linux, you can still load the driver with\n"));
                        CAN_PRINT(("esd CAN driver: module parameter \"compat32=0\", but you won't be\n"));
                        CAN_PRINT(("esd CAN driver: able to use 32-bit applications.\n"));
                        esdcan_unregister_ioctl32();
                }
        }
        return result;
}
#else
int32_t esdcan_register_ioctl32(void) {return 0;}
void esdcan_unregister_ioctl32(void) {}
#endif

/*
 *  The following includes contain system-specific code, which helps to reduce code
 *  duplication and ifdefs for different hardware.
 */
#if defined (BOARD_USB)
# if defined(OSIF_OS_LINUX)
#  include "esdcan_usb.c"
# else
#  error "Currently USB is implemented for Linux, only!"
# endif
#elif defined(ESDCAN_PCI_HOTPLUG)
# include "esdcan_pci.c"
#elif defined(BOARD_RAW)
# include "esdcan_raw.c"
#elif defined(BOARD_PCI)
# include "esdcan_pci_legacy.c"
#else
# error "Board type not properly defined!!!"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static irqreturn_t irq_stub( int irq, void *arg, struct pt_regs* regs )
{
        return ((CARD_IRQ*)arg)->handler(((CARD_IRQ*)arg)->context);
}
#else
static void irq_stub( int irq, void *arg, struct pt_regs* regs )
{
        ((CARD_IRQ*)arg)->handler(((CARD_IRQ*)arg)->context);
        return;
}
#endif

#include "esdcan_common.c" /* include common part of unix system layer */

#ifdef OSIF_OS_LINUX
static int esdcan_open(struct inode *inode, struct file *file)
{
	CAN_DBG((ESDCAN_ZONE_FU, "%s: pre LOCK\n", OSIF_FUNCTION));
	HOTPLUG_GLOBAL_LOCK;

# if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	MOD_INC_USE_COUNT;
# endif
	CAN_DBG((ESDCAN_ZONE_FU, "%s: enter\n", OSIF_FUNCTION));

	file->private_data = 0;

	HOTPLUG_GLOBAL_UNLOCK;

	CAN_DBG((ESDCAN_ZONE_FU, "%s: leave\n", OSIF_FUNCTION));

	RETURN_TO_USER(OSIF_SUCCESS);
}
#endif


#ifdef OSIF_OS_LINUX
static int esdcan_release(struct inode *inode, struct file *file)
{
	INT32 result = OSIF_SUCCESS;
	CAN_OCB *ocb = (CAN_OCB*)file->private_data;
	CAN_NODE *node;

	CAN_DBG((ESDCAN_ZONE_FU, "%s: enter\n", OSIF_FUNCTION));

	if (ocb) {
                node = ocb->node;

		if ( 0 != HOTPLUG_BOLT_USER_ENTRY( ((CAN_CARD*)node->crd)->p_mod ) ) { /* AB: seems no good, needs rework */
			CAN_DBG((ESDCAN_ZONE_FU, "%s: release needs abort!!!\n", OSIF_FUNCTION));
			RETURN_TO_USER(result);
		}

		OSIF_MUTEX_LOCK(&node->lock);
		result = ocb_destroy(ocb);
		file->private_data = 0;
		OSIF_MUTEX_UNLOCK(&node->lock);

		HOTPLUG_BOLT_USER_EXIT( ((CAN_CARD*)node->crd)->p_mod );
	}

# if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	MOD_DEC_USE_COUNT;
# endif

	CAN_DBG((ESDCAN_ZONE_FU, "%s: leave\n", OSIF_FUNCTION));
	RETURN_TO_USER(result);
}
#endif /* OSIF_OS_LINUX */


INT32 can_attach_common(CAN_CARD *crd)
{
	VOID       *vptr;
        INT32       nodesUsed;
        INT32       idx_node, idx_irq;
        INT32       result = OSIF_SUCCESS, lastError = OSIF_SUCCESS;
        CARD_IDENT *pCardIdent = crd->pCardIdent;

        CAN_DBG((ESDCAN_ZONE_INI, "%s: enter (crd=%p)\n", OSIF_FUNCTION, crd));

        for(idx_node = 0; idx_node < NODES_PER_CARD; idx_node++) {
                CAN_NODE *node;

                result = OSIF_MALLOC(sizeof(*node), &vptr);
                if(OSIF_SUCCESS != result) {
                        break;
                }
                node = (CAN_NODE*)vptr;
                OSIF_MEMSET(node, 0, sizeof(*node));

                node->crd     = crd;
                node->mode    = mode;  /* insmod parameter */
                node->net_no  = crd->card_no*NODES_PER_CARD + idx_node;
                node->node_no = idx_node;

                OSIF_MUTEX_CREATE(&node->lock);
                OSIF_IRQ_MUTEX_CREATE(&node->lock_irq);
                crd->node[idx_node] = node;
        }
        if( OSIF_SUCCESS != result ) {
                lastError = result;
                goto common_attach_1;
        }

        result = can_board_attach(crd);
        if( OSIF_SUCCESS != result ) {
                lastError = result;
                goto common_attach_1;   /* AB TODO: this leads to an undestroyed crd->lock_irq !!! */
        }

        for(idx_irq = 0;
            (idx_irq < 8) &&
            (pCardIdent->all.irqs[idx_irq].handler != (VOID*)(~0x0));
            idx_irq++) {
                if (NULL == pCardIdent->all.irqs[idx_irq].handler) {
                        continue;
                }
/* AB TODO: timestamps on baudrate change events
Idea: force affinity to the same CPU as DPC thread
set_ioapic_affinity
irq_desc[irq].handler->set_affinity(irq, new_value)
UINT32 cpus = num_online_cpus();
cpumask_t mask;
cpus_clear(mask);
cpus -= 1;
cpu_set(cpus, mask);
long sched_setaffinity(pid_t pid, cpumask_t new_mask)
 */
                if (0 != request_irq(crd->irq[idx_irq],
                                     (void*)irq_stub,
                                     SA_SHIRQ | SA_INTERRUPT,
                                     pCardIdent->all.name,
                                     &crd->irqs[idx_irq])) {
                        result = CANIO_EBUSY;
                        CAN_DBG((ESDCAN_ZONE_INI,
                                 "%s: IRQ request FAILURE: IRQ: %d, Stub: %p, Param: %p\n",
                                 OSIF_FUNCTION,
                                 crd->irq[idx_irq],
                                 (void*)irq_stub,
                                 &crd->irqs[idx_irq]));
                        break;
                }
                CAN_DBG((ESDCAN_ZONE_INI, "%s: irq requested: IRQ: %d, Stub: %p, Param: %p\n",
                         OSIF_FUNCTION,
                         crd->irq[idx_irq],
                         (void*)irq_stub,
                         &crd->irqs[idx_irq]));

                crd->irqs[idx_irq] = pCardIdent->all.irqs[idx_irq];
        }
        if( OSIF_SUCCESS != result ) {
                lastError = result;
                goto common_attach_2;
        }

        nodesUsed = 0;
        for( idx_node = 0; idx_node < NODES_PER_CARD; idx_node++ ) {
                CAN_NODE *node = crd->node[idx_node];
                if(NULL == node) {
                        continue;
                }

                node->features |= crd->features;

                CAN_DBG((ESDCAN_ZONE_INI, "%s: calling nuc_node_attach\n", OSIF_FUNCTION));
                result = nuc_node_attach( node );
                if (OSIF_SUCCESS != result) {
                        CAN_DBG((ESDCAN_ZONE_INI, "%s: nuc_node_attach failed (%d)\n", OSIF_FUNCTION, result));
                        OSIF_IRQ_MUTEX_DESTROY( &node->lock_irq );
                        OSIF_MUTEX_DESTROY( &node->lock );
                        OSIF_FREE( node );

                        crd->node[idx_node] = NULL;
                }

                nodes[crd->card_no*NODES_PER_CARD + idx_node] = crd->node[idx_node];
                if ( NULL == crd->node[idx_node]) {
                        continue;
                }

                CAN_DBG((ESDCAN_ZONE_INI, "%s: Node attached (features=%08x)\n", OSIF_FUNCTION, node->features));
                nodesUsed++;
        }
        if( 0 == nodesUsed ) {
                CAN_DBG((ESDCAN_ZONE_INI, "%s: no node working\n", OSIF_FUNCTION));
                lastError = result;
                goto common_attach_3;
        }

        result = can_board_attach_final(crd);
        if( OSIF_SUCCESS != result ) {
                lastError = result;
                goto common_attach_3;
        }

        CAN_DBG((ESDCAN_ZONE_INI, "%s: leave (result=%d)\n", OSIF_FUNCTION, result));
        return result;

 common_attach_3:
        CAN_DBG((ESDCAN_ZONE_INI,
                 "%s: common_attach_3: can_board_attach_final or nuc_node_attach failed\n",
                 OSIF_FUNCTION));
        for( idx_node = 0; idx_node < NODES_PER_CARD; idx_node++ ) {
                CAN_NODE *node = crd->node[idx_node];
                if ( NULL == node ) {
                        continue;
                }

                nodes[crd->card_no * NODES_PER_CARD + idx_node] = NULL;

                nuc_node_detach( node );
        }

 common_attach_2:
        CAN_DBG((ESDCAN_ZONE_INI,
                 "%s: common_attach_2: Failed to request IRQ\n",
                 OSIF_FUNCTION));
        for( idx_irq = 0;
             (idx_irq < 8) &&
             (pCardIdent->all.irqs[idx_irq].handler != (VOID*)(-1));
             idx_irq++ ) {
                if( NULL != crd->irqs[idx_irq].context ) {
                        free_irq(crd->irq[idx_irq], &crd->irqs[idx_irq]);
                }
        }

        can_board_detach(crd);

 common_attach_1:
        CAN_DBG((ESDCAN_ZONE_INI,
                 "%s: common_attach_1: Allocation of node-struct or can_board_attach failed\n",
                 OSIF_FUNCTION));
        for( idx_node-- /* current idx did't get the mem */;
             idx_node >= 0;
             idx_node--) {
                CAN_NODE *node = crd->node[idx_node];
                if ( NULL == node ) {
                        continue;
                }
                OSIF_IRQ_MUTEX_DESTROY(&node->lock_irq);
                OSIF_MUTEX_DESTROY(&node->lock);
                OSIF_FREE(node);
        }

        CAN_DBG((ESDCAN_ZONE_INI, "%s: leave (error=%d)\n", OSIF_FUNCTION, lastError));
        return lastError;
}


VOID can_detach_common( CAN_CARD *crd, INT32 crd_no )
{
	UINT32 j;
        CARD_IDENT *pCardIdent = crd->pCardIdent;

	can_board_detach(crd);

	for ( j = 0; (j < 8) && ((VOID *)(~0x0) != pCardIdent->all.irqs[j].handler); j++ ) {
		if ( NULL != crd->irqs[j].context ) {
                        free_irq(crd->irq[j], &crd->irqs[j]);
		}
	}

	for ( j = 0; j < NODES_PER_CARD; j++ ) {
		CAN_NODE *node = crd->node[j];
		if ( NULL == node ) {
			continue;
		}

		nodes[crd_no * NODES_PER_CARD + j] = NULL;

		nuc_node_detach( node );
		OSIF_IRQ_MUTEX_DESTROY( &node->lock_irq);
		OSIF_MUTEX_DESTROY( &node->lock);
		OSIF_FREE(node);
	}

#ifdef ESDCAN_PCI_HOTPLUG
        for ( j = 0; j < 8; j++ ) {
                if ( 0 != crd->range[j] ) {
                        if ( pci_resource_flags(crd->pciDev, j) & IORESOURCE_MEM ) {
                                /* Memory space */
                                iounmap(crd->base[j]);
                        } else if ( pci_resource_flags(crd->pciDev, j) & IORESOURCE_IO ) {
                                /* IO-space */
                                release_region((UINTPTR)crd->base[j], crd->range[j]);
                        }
                }
        }
#else
	for ( j = 0; (j < 8) && (0xFFFFFFFF != pCardIdent->all.spaces[j]); j++ ) {
		if( 0 != (pCardIdent->all.spaces[j] & 0x00000001) ) { /* if io-space */
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			release_region((UINTPTR)crd->base[j], crd->range[j]);
# else
#  warning "AB TODO: Still need mechanism to release mem-regions in 2.2.x kernels!!!"
# endif
		} else {
			iounmap(crd->base[j]);
		}
	}
#endif
        return;
}


VOID esdcan_show_card_info( CAN_CARD *crd )
{
        INT32 i;

        if ( (crd->version_firmware.level != 0) ||
             (crd->version_firmware.revision != 0) ||
             (crd->version_firmware.build != 0) ) {
                CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                            "esd CAN driver: Firmware-version = %1x.%1x.%02x\n",
                            crd->version_firmware.level,
                            crd->version_firmware.revision,
                            crd->version_firmware.build));
        }
        if ( (crd->version_hardware.level != 0) ||
             (crd->version_hardware.revision != 0) ||
             (crd->version_hardware.build != 0) ) {
			CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                                    "esd CAN driver: Hardware-version = %1x.%1x.%02x\n",
                                    crd->version_hardware.level,
                                    crd->version_hardware.revision,
                                    crd->version_hardware.build));
        } else {
                CAN_DPRINT((ESDCAN_ZONE_USR_INIT, "esd CAN driver: Hardware-version = 1.0.00\n"));
        }

        CAN_DPRINT((ESDCAN_ZONE_USR_INIT, "esd CAN driver: Card = %d Minor(s) =", crd->card_no));
        for ( i = 0; NODES_PER_CARD > i; i++ ) {
                if ( NULL == crd->node[i] ) {
                        continue;
                }

                CAN_DPRINT((ESDCAN_ZONE_USR_INIT, " %d,", crd->card_no * NODES_PER_CARD + i));
        }
        return;
}


/*!
 *  \brief This function prints all information, that might be useful to the user
 *  at the time the driver is loaded. Use this function in init-functions (such as
 *  esdcan_init())
 */
VOID esdcan_show_driver_info( VOID )
{
	CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                    "esd CAN driver: %s\n", ESDCAN_DRIVER_NAME));

	if ( baud != CANIO_BAUD_NOT_SET ) {
		CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                            "esd CAN driver: initial baudrate = %d\n", baud));
	} else {
		CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                            "esd CAN driver: baudrate not set\n"));
	}

	CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                    "esd CAN driver: mode = 0x%08x, major = %d, verbose = 0x%08x\n",
                    mode, major, verbose));

	CAN_DPRINT((ESDCAN_ZONE_USR_INIT,
                    "esd CAN driver: version %d.%d.%d "__TIME__" "__DATE__": successfully loaded\n",
                    LEVEL, REVI, BUILD));

#if defined(CAN_DEBUG)
        CAN_PRINT(("esd CAN driver: !!! DEBUG VERSION !!!\n"));
        CAN_PRINT(("Please note:\n"));
        CAN_PRINT(("You're using a debug version of the esd CAN driver.\n"));
        CAN_PRINT(("This version is NOT intended for productive use.\n"));
        CAN_PRINT(("The activated debug code might have bad influence\n"));
        CAN_PRINT(("on performance and system load.\n"));
        CAN_PRINT(("If you received this driver from somebody else than\n"));
        CAN_PRINT(("esd support, please report to \"support@esd-electronics.com\".\n"));
        CAN_PRINT(("Thank you!\n"));
#endif
        return;
}


#ifndef ESDCAN_HOTPLUG_DRIVER
VOID esdcan_show_card_info_all( VOID )
{
	UINT32 i;

	for ( i = 0; i < MAX_CARDS; i++ ) {
		CAN_CARD *crd;

		HOTPLUG_BOLT_SYSTEM_ENTRY(modules[i]);

		crd = GET_CARD_POINTER(i);
		if ( NULL == crd ) {
			HOTPLUG_BOLT_SYSTEM_EXIT(modules[i]);
			continue;
		}

                esdcan_show_card_info(crd);

		HOTPLUG_BOLT_SYSTEM_EXIT(modules[i]);

		CAN_DPRINT((ESDCAN_ZONE_USR_INIT, "\n"));
	}
        return;
}

INT32 can_detach_legacy( VOID )
{
	INT32 result = OSIF_SUCCESS;
	UINT32 i;

	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: enter\n", OSIF_FUNCTION));

	for( i = 0; i < MAX_CARDS; i ++) {
		CAN_CARD *crd = GET_CARD_POINTER(i);
		if ( NULL == crd ) {
			continue;
		}

		GET_CARD_POINTER(i) = NULL;

		can_detach_common(crd, i);

		OSIF_FREE( crd );
	}

	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: leave\n", OSIF_FUNCTION));

	return result;
}

/*! Executed when driver module is loaded */
int OSIF__INIT esdcan_init_legacy(void)
{
        INT32 result = 0;

	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: enter\n", OSIF_FUNCTION));

# ifdef LTT
	can_ltt0 = trace_create_event("CAN_LTT_0",
				      NULL,
				      CUSTOM_EVENT_FORMAT_TYPE_HEX,
				      NULL);
# endif /* LTT */

        if (esdcan_register_ioctl32()) {
                RETURN_TO_USER(OSIF_INVALID_PARAMETER);
        }

	if ( OSIF_ATTACH() ) {
		CAN_DBG((ESDCAN_ZONE_INI, "%s: osif_attach failed!!!\n", OSIF_FUNCTION));
                esdcan_unregister_ioctl32();
                RETURN_TO_USER(OSIF_INSUFFICIENT_RESOURCES);
        }

	result = can_attach();
	if ( result != OSIF_SUCCESS ) {
		OSIF_DETACH();
                esdcan_unregister_ioctl32();
 		RETURN_TO_USER(result);
	}

	esdcan_show_driver_info();
	esdcan_show_card_info_all();

# ifndef OSIF_OS_RTAI
	/* register device */
	if ( 0 != register_chrdev( major, ESDCAN_DRIVER_NAME, &esdcan_fops ) ) {
		CAN_PRINT(("esd CAN driver: cannot register character device for major=%d\n", major));
                CAN_DETACH;
		OSIF_DETACH();
                esdcan_unregister_ioctl32();
		RETURN_TO_USER(EBUSY);
	}
# endif /* OSIF_OS_RTAI */

        esdcan_proc_bus_statistic_create();
        esdcan_proc_extra_info_create();

	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: leave\n", OSIF_FUNCTION));
	RETURN_TO_USER(result);
}

/*! Executed when driver module is unloaded */
void OSIF__EXIT esdcan_exit_legacy(void)
{
	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: enter, unloading esd CAN driver\n", OSIF_FUNCTION));

	unregister_chrdev(major, ESDCAN_DRIVER_NAME);

	esdcan_unregister_ioctl32();

	CAN_DETACH;

# ifdef LTT
	trace_destroy_event(can_ltt0);
# endif

        esdcan_proc_bus_statistic_remove();
        esdcan_proc_extra_info_remove();

	OSIF_DETACH();

	CAN_DPRINT((ESDCAN_ZONE_USR_INIT, "esd CAN driver: unloaded\n"));

	CAN_DBG((ESDCAN_ZONE_INIFU, "%s: leave\n", OSIF_FUNCTION));
}

OSIF_MODULE_INIT(esdcan_init_legacy);
OSIF_MODULE_EXIT(esdcan_exit_legacy);
#endif /* !defined(ESDCAN_HOTPLUG_DRIVER) */

MODULE_AUTHOR("esd GmbH");
MODULE_DESCRIPTION("esd CAN driver");

#ifdef OSIF_OS_RTAI
EXPORT_SYMBOL(esdcan_read);
EXPORT_SYMBOL(esdcan_write);
EXPORT_SYMBOL(esdcan_ctrl);
#else
# if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
# endif
#endif


/* Id for RCS version system */
#if ((__GNUC__ > 2) || (__GNUC__ == 2 && __GNUC_MINOR__ >=7))
static char* rcsid __attribute__((unused)) = "$Id: esdcan.c,v 1.109 2006/10/12 09:53:36 andreas Exp $";
#else  /* No or old GNU compiler */
# define USE(var) static void use_##var(void *x) {if(x) use_##var((void *)var);}
static char* rcsid = "$Id: esdcan.c,v 1.109 2006/10/12 09:53:36 andreas Exp $";
USE(rcsid);
#endif /* of GNU compiler */
